import sys
if sys.version_info > (3, 3):
    from inspect import signature

    def pkind(p):
        return p.kind.name
else:
    from funcsigs import signature

    def pkind(p):
        return p.kind._name


def infer_title(task):
    if hasattr(task, "__name__"):
        if hasattr(task, "__module__"):
            prefix = getattr(task, "__module__") + "."
        else:
            prefix = ""
        return prefix + task.__name__
    else:
        raise ValueError("The title of %r could be infered." % task)


class CallMap(object):

    def __init__(self, function):
        self.parameters = dict(signature(function).parameters)

    def map(self, args, kwargs):
        mapping, used = {}, []
        for i, name in enumerate(self.parameters):
            param = self.parameters[name]
            handler = getattr(self, pkind(param))
            value = handler(i, param, args, kwargs, used)
            if value is not param.empty:
                mapping[param.name] = value
        return mapping

    @staticmethod
    def POSITIONAL_ONLY(index, param, args, kwargs, used_keys):
        return args[index]

    @staticmethod
    def POSITIONAL_OR_KEYWORD(index, param, args, kwargs, used_keys):
        used_keys.append(param.name)
        if len(args) > index:
            return args[index]
        elif param.name in kwargs:
            return kwargs[param.name]
        else:
            return param.default

    @staticmethod
    def VAR_POSITIONAL(index, param, args, kwargs, used_keys):
        return list(args[index:])

    @staticmethod
    def KEYWORD_ONLY(index, param, args, kwargs, used_keys):
        return kwargs.get(param.name, param.default)

    @staticmethod
    def VAR_KEYWORD(index, param, args, kwargs, used_keys):
        return {k : v for k, v in kwargs.items() if k in used_keys}



import sys
import json
import datetime
import inspect
from copy import deepcopy

from .util import infer_title


class Style(object):

    default = None

    def __call__(self, log):
        log = deepcopy(log)
        method = getattr(self, log["metadata"]["status"], self.default)
        if method is not None:
            log = self._pre(log)
            if log:
                lines = method(log)
                if lines and isinstance(lines, str):
                    lines = [lines]
                else:
                    lines = list(lines)
                if lines:
                    return self._post("\n".join(lines) + "\n")

    def _pre(self, log):
        return log

    def _post(self, log):
        return log


class Raw(Style):

    class Encoder(json.JSONEncoder):

        def default(self, o):
            try:
                return super(json.JSONEncoder, self).default(o)
            except:
                return str(o)

    def default(self, log):
        return self.Encoder().encode(log)


class Log(Style):

    def _pre(self, log):
        if isinstance(log["metadata"]["title"], str):
            return log

    def _completed(self, log):
        metadata = log["metadata"]
        status = metadata["status"]
        duration = metadata["stop"] - metadata["start"]
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        info = ", ".join(map(lambda i: "%s: %s" % i, log["content"].items()))
        message = "{time} {status} {title} after {duration:.3f} seconds"
        if len(info) > 50:
            info = info[:47] + "..."
        if info:
            message += " - {info}"
        yield message.format(time=timestamp, status=status.upper(),
            title=metadata["title"], info=info, duration=duration)

    success = _completed
    failure = _completed


class Tree(Style):

    def _pre(self, log):
        metadata = log["metadata"]
        if not isinstance(metadata["title"], str):
            try:
                metadata["title"] = infer_title(metadata["title"])
            except:
                pass
        return log

    def started(self, log):
        content, metadata = log["content"], log["metadata"]
        indent = "|   " * metadata["depth"]
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        yield indent + "|-- {status}: {title}".format(**metadata)
        yield indent + "|   @ {0}".format(timestamp)
        if "reason" in content:
            yield indent + "|   | reason: {0}".format(content.pop("reason"))
        for k, v in content.items():
            yield indent + "|   | {0}: {1}".format(k, v)

    def working(self, log):
        content, metadata = log["content"], log["metadata"]
        indent = "|   " * (metadata["depth"] + 1)
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        yield indent + "|-- {status}: {title}".format(**metadata)
        yield indent + "|   @ {0}".format(timestamp)
        if "reason" in content:
            yield indent + "|   | reason: {0}".format(content.pop("reason"))
        for k, v in content.items():
            yield indent + "|   | {0}: {1}".format(k, v)

    def success(self, log):
        content, metadata = log["content"], log["metadata"]
        indent = "|   " * (metadata["depth"] + 1)
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        yield indent + "`-- {status}: {title}".format(**metadata)
        yield indent + "    @ {0}".format(timestamp)
        diff = metadata["stop"] - metadata["start"]
        content["duration"] = "{:.3f}".format(diff) + " seconds"
        if "reason" in content:
            yield indent + "|   | reason: {0}".format(content.pop("reason"))
        for k, v in content.items():
            yield indent + "    | {0}: {1}".format(k, v)

    def failure(self, log):
        content, metadata = log["content"], log["metadata"]
        indent = "|   " * (metadata["depth"] + 1)
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        yield indent + "`-- {status}: {title}".format(**metadata)
        yield indent + "    @ {0}".format(timestamp)
        diff = metadata["stop"] - metadata["start"]
        content["duration"] = "{:.3f}".format(diff) + " seconds"
        if "reason" in content:
            yield indent + "|   | reason: {0}".format(content.pop("reason"))
        for k, v in content.items():
            yield indent + "    | {0}: {1}".format(k, v)



import os
import sys
import json
from time import time as now
from multiprocessing import (
    Process, JoinableQueue as ProcessQueue, Event as ProcessEvent)
if sys.version_info > (3, 0):
    from queue import Queue as ThreadQueue
else:
    from Queue import Queue as ThreadQueue
from threading import Thread, Event as ThreadEvent


class DistributorMixin(object):

    def __init__(self, inbox, stop):
        super(DistributorMixin, self).__init__()
        for attr in ("join", "get", "put", "task_done"):
            if not hasattr(inbox, attr):
                raise TypeError("Expected some type of joinable"
                    " queue that implements %r." % attr)
        self.inbox = inbox
        self.daemon = True
        self._outlets = []
        self._stop = stop

    def set_outlets(self, *outlets):
        self._outlets = list(outlets)

    def __call__(self, log):
        if not self.is_alive():
            self.start()
        self.inbox.put(log)

    def deadline(self, timeout):
        start = now()
        while not self.inbox.empty():
            if not self.is_alive():
                self.start()
            if now() - start > timeout:
                self.stop()
                break

    def stop(self):
        self._stop.set()

    def run(self):
        while not self._stop.is_set():
            log = self.inbox.get()
            try:
                self.send(log)
            except:
                raise
            finally:
                self.inbox.task_done()

    def send(self, log):
        for o in self._outlets:
            o(log)


class DistributorProcess(DistributorMixin, Process):

    def __init__(self):
        super(DistributorProcess, self).__init__(ProcessQueue(), ProcessEvent())


class DistributorThread(DistributorMixin, Thread):

    def __init__(self):
        super(DistributorThread, self).__init__(ThreadQueue(), ThreadEvent())
